IMPLEMENTATION MODULE Client;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Str, Types, Lib2, Lib3, SO, Crc, MD5, Tmrs, Tmr, UFTP, Cfg, Log, Config, SrvIO, Files, Rcvrs, Sndrs;

CONST
  mid = Cfg.midClient;

  dtRnd  = 20;          (* ms, round-trip time (estimated) *)
  toSes  = 50 * dtRnd;  (* ms, session request period *)
  toDat  =  5 * dtRnd;  (* ms, data request period *)
  fglMax = UFTP.lenMaxFgm;

TYPE
  TState = ( stOpen, stData, stClose );

  TSesDesc = RECORD
    st: TState;
    t : Tmr.TTicks;
    f : Files.TFile;
    CASE w (* wrt *): BOOLEAN OF
    | FALSE: s: Sndrs.TSndr;
    | TRUE : r: Rcvrs.TRcvr;
    END;
    rf: UFTP.TReference;
  END;

VAR
  ses: TSesDesc;
  err: TErrCnts;
  tmr: Tmrs.TTimer;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* solve return code problem before removing this line *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrTgt( VAR b: SO.TBuffer; t: UFTP.TPosDesc );
VAR
  c: CHAR;
  s: MD5.TDgstStr;
BEGIN
  c := b.c.p;
  b.c.p := '0';
  SO.WrLngHex( b, t.l, 8 );
  SO.WrChar( b, ' ' );
  MD5.MkStr( s, t.d );
  SO.WrStr( b, s );
  b.c.p := c;
END WrTgt;


PROCEDURE ShowErrCnts();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
  e: TErr;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'ShowErrCnts' );
  FOR e := MIN(TErr) TO MAX(TErr) DO
    SO.WrChar( b, ' ' );
    SO.WrLngHex( b, err[e] MOD 1000H, 3 );
  END;
  Message( s );
END ShowErrCnts;


PROCEDURE IncErrCnt( e: TErr );
BEGIN
  Lib2.IncL( err[e], 1 );
  Lib2.IncL( err[eNull], 1 );
END IncErrCnt;

PROCEDURE OnRx( dat-: ARRAY OF BYTE; len: CARDINAL );
VAR
  tel: POINTER TO UFTP.TTelDesc;

  PROCEDURE RdRp();
  CONST
    lenMin = SIZE( UFTP.TRdRpDesc );
    lenMax = lenMin;
  VAR
    f: UFTP.TFgmLen;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..127] OF CHAR;
    BEGIN
      IF ~ ena THEN RETURN END;
      SO.Init( b, s );
      SO.WrStr( b, 'RdRp fgl ' );
      SO.WrLngCard( b, tel^.rrp.f, 1 );
      SO.WrStr( b, ' tgt ' );
      WrTgt( b, tel^.rrp.p );
      SO.WrStr( b, ' ref ' );
      SO.WrLngCard( b, tel^.rrp.r, 1 );
      Message( s );
    END Show;

  BEGIN
    IF ses.st # stOpen THEN
      IncErrCnt( eRdRpSt );
      RETURN;
    END;
    IF ses.w THEN
      IncErrCnt( eRdRpWr );
      RETURN;
    END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eRdRpLen );
      RETURN;
    END;
    CASE tel^.rrp.e OF
    | UFTP.errNull: (* ok *)
    | UFTP.errRID : Fatal( 'Resource not found.' );
    | UFTP.errRes : Fatal( 'Server full.' );
    ELSE
      Fatal( 'Unknown error.' );
    END;
    f := tel^.rrp.f;
    IF ( f = 0 ) OR ( f > fglMax ) THEN
      IncErrCnt( eRdRpFgl );
      RETURN;
    END;
    ses.t := Tmr.Ticks();
    Files.SetTgt( ses.f, tel^.rrp.p );
    (* wnd = fgf * dtr (fragment frequency times round-trip time) *)
    (* fgf = fpu / dtu *)
    ses.r := Rcvrs.Alloc( f, ses.f, Config.data.fpu * dtRnd DIV Config.data.dtu );
    ses.rf := tel^.rrp.r;
    ses.st := stData;
    Show();
  END RdRp;

  PROCEDURE DatRp();
  CONST
    lenMin = UFTP.lenMinDatRpDesc;
    lenMax = UFTP.lenMaxDatRpDesc;

    PROCEDURE Show();
    CONST
      ena = FALSE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..511] OF CHAR;
    BEGIN
      IF ~ ena THEN RETURN END;
      SO.Init( b, s );
      SO.WrStr( b, 'DatRp idx ' );
      SO.WrLngCard( b, tel^.drp.h.i, 1 );
      SO.WrStr( b, ' len ' );
      SO.WrLngCard( b, len, 1 );
      Message( s );
    END Show;

  BEGIN
    IF ses.st # stData THEN
      IncErrCnt( eDatRpSt );
      RETURN
    END;
    IF ses.w THEN
      IncErrCnt( eDatRpWr );
      RETURN;
    END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRpLen );
      RETURN;
    END;
    DEC( len, SIZE( UFTP.TDatHdrDesc ) );
    Show();
    Rcvrs.Receive( ses.r, tel^.drp.h.i, tel^.drp.d, len );
  END DatRp;

CONST
  lenMin = UFTP.lenFrmOvr;
  lenMax = UFTP.lenMaxTel;
BEGIN
  IF ( len < lenMin ) OR ( len > lenMax ) THEN IncErrCnt( eLen ); RETURN END;
  IF ~ Crc.Ok( dat, len ) THEN IncErrCnt( eCrc ); RETURN END;
  tel := ADR( dat );
  DEC( len, lenMin );
  IF tel^.hdr.m # UFTP.mag THEN IncErrCnt( eMag ); RETURN END;
  IF tel^.hdr.v # UFTP.ver THEN IncErrCnt( eVer ); RETURN END;
  CASE tel^.hdr.c OF
  | UFTP.cmdRdRp : RdRp();
  | UFTP.cmdDatRp: DatRp();
  ELSE
    IncErrCnt( eCmd );
  END;
END OnRx;


PROCEDURE DoOpen();

  PROCEDURE RdRq();
  VAR
    tel: RECORD h: UFTP.THdrDesc; b: UFTP.TRdRqDesc; t: UFTP.TTrlDesc END;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..63] OF CHAR;
    BEGIN
      IF ~ ena THEN RETURN END;
      SO.Init( b, s );
      SO.WrStr( b, 'RdRq fgl ' );
      SO.WrCard( b, tel.b.f, 1 );
      SO.WrStr( b, " rid '" );
      SO.WrStr( b, tel.b.r );
      SO.WrStr( b, "'" );
      Message( s );
    END Show;

  BEGIN
    Lib2.Fill( tel, 0 );
    tel.h.m := UFTP.mag;
    tel.h.v := UFTP.ver;
    tel.h.c := UFTP.cmdRdRq;
    tel.b.f := fglMax;
    Files.GetRID( ses.f, tel.b.r );
    tel.t := Crc.Swap( Crc.UpdL( Crc.crcInit, tel, SIZE(tel) - SIZE(tel.t) ) );
    SrvIO.Send( tel, SIZE( tel ) );
    Show();
  END RdRq;

  PROCEDURE WrRq();
  VAR
    tel: RECORD h: UFTP.THdrDesc; b: UFTP.TWrRqDesc; t: UFTP.TTrlDesc END;
  BEGIN
    (* !!! *)
  END WrRq;

VAR
  t: Tmr.TTicks;
BEGIN
  t := Tmr.Ticks();
  IF Tmr.Diff( ses.t, t ) < toSes THEN RETURN END;
  ses.t := t;
  IF ses.w THEN WrRq() ELSE RdRq() END;
END DoOpen;

PROCEDURE DatRq( r: Rcvrs.TRef; i: UFTP.TFgmIdx; s: UFTP.TFgmSet );
VAR
  tel: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRqDesc; t: UFTP.TTrlDesc END;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DatRq idx ' );
    SO.WrLngCard( b, tel.b.h.i, 1 );
    SO.WrStr( b, ' set ' );
    SO.WrLngHex( b, LONGCARD( tel.b.s ), 8 );
    Message( s );
  END Show;

BEGIN
  Lib2.Fill( tel, 0 );
  tel.h.m := UFTP.mag;
  tel.h.v := UFTP.ver;
  tel.h.c := UFTP.cmdDatRq;
  tel.b.h.r := ses.rf;
  tel.b.h.i := i;
  tel.b.s := s;
  tel.t := Crc.Swap( Crc.UpdL( Crc.crcInit, tel, SIZE(tel) - SIZE(tel.t) ) );
  SrvIO.Send( tel, SIZE( tel ) );
  Show();
END DatRq;

PROCEDURE DoData();
BEGIN
  IF ses.w THEN
  ELSE
    IF ~ Rcvrs.Update( ses.r, Config.data.fpu, toDat DIV Config.data.dtu, DatRq, 0 ) THEN ses.st := stClose END;
  END;
END DoData;

PROCEDURE DoClose();
VAR
  tel: RECORD h: UFTP.THdrDesc; b: UFTP.TCloseDesc; t: UFTP.TTrlDesc END;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
    dt: Tmr.TTicks;
  BEGIN
    IF ~ ena THEN RETURN END;
    dt := Tmr.Ticks() - ses.t;
    SO.Init( b, s );
    SO.WrStr( b, 'DoClose ' );
    SO.WrLngCard( b, dt, 1 );
    SO.WrStr( b, ' ms' );
    Message( s );
  END Show;

VAR
  t, p: UFTP.TPosDesc;
BEGIN
  Show();
  IF ses.w THEN
  ELSE
    Rcvrs.Free( ses.r );
    Files.GetTgt( ses.f, t );
    Files.GetPos( ses.f, 0, p );
    IF Lib2.Compare( t, p ) # SIZE( p ) THEN Fatal( 'DoClose tgt mismatch' ) END;
    Files.Close( ses.f );
  END;

  Lib2.Fill( tel, 0 );
  tel.h.m := UFTP.mag;
  tel.h.v := UFTP.ver;
  tel.h.c := UFTP.cmdClose;
  tel.b.r := ses.rf;
  tel.t := Crc.Swap( Crc.UpdL( Crc.crcInit, tel, SIZE(tel) - SIZE(tel.t) ) );
  SrvIO.Send( tel, SIZE( tel ) );

  Lib3.SetReturnCode( 0 );
  HALT;
END DoClose;

PROCEDURE OnTmr( k: Tmrs.TTicks; r: Tmrs.TRef );
BEGIN
  CASE ses.st OF
  | stOpen : DoOpen();
  | stData : DoData();
  | stClose: DoClose();
  ELSE
    Fatal( 'OnTmr Invalid state' );
  END;
END OnTmr;


PROCEDURE Run();
BEGIN
  SrvIO.Execute();
END Run;

PROCEDURE GetErrCnts( VAR e: TErrCnts );
BEGIN
  e := err;
END GetErrCnts;


PROCEDURE InitSes();
VAR
  r: UFTP.TResID;
BEGIN
  Lib2.Fill( ses, 0 );
  ses.w := Config.data.wrt;
  ses.t := Tmr.Sub( Tmr.Ticks(), toSes );  (* force initial time-out *)
  Str.Copy( r, Config.data.rnm );
  IF ses.w THEN
    ses.f := Files.OpenRd( r );
  ELSE
    ses.f := Files.OpenWr( r );
  END;
END InitSes;

PROCEDURE DoneSes();
BEGIN
  IF ses.f # Files.fNull THEN Files.Close( ses.f ) END;
END DoneSes;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
BEGIN
  DoneSes();
  ShowErrCnts();
  Message( 'DoneModule' );
  oldTerm();
END DoneModule;

PROCEDURE InitModule();

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'InitModule UFTP ver ' );
    SO.WrCard( b, UFTP.ver, 1 );
    SO.WrStr( b, ' lenMaxTel ' );
    SO.WrCard( b, UFTP.lenMaxTel, 1 );
    SO.WrStr( b, ' lenFrmOvr ' );
    SO.WrCard( b, UFTP.lenFrmOvr, 1 );
    SO.WrStr( b, ' lenMaxVar ' );
    SO.WrCard( b, UFTP.lenMaxVar, 1 );
    SO.WrStr( b, ' lenMaxFgm ' );
    SO.WrCard( b, UFTP.lenMaxFgm, 1 );
    Message( s );
  END Show;

BEGIN
  Log.Register( mid, 'Client' );
  Lib2.Fill( ses, 0 );
  Lib2.Fill( err, 0 );

  tmr := Tmrs.New();
  Tmrs.Set( tmr, 1, Config.data.dtu );
  Tmrs.SetHnd( tmr, OnTmr, 0 );

  InitSes();

  SrvIO.Install( OnRx );

  Lib3.Terminate( DoneModule, oldTerm );

  Show();
END InitModule;

BEGIN
  InitModule();
END Client.
