IMPLEMENTATION MODULE Files;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib2, Lib3, Str, FIO, FIOStd;
IMPORT Cfg, MD5, UFTP, SO, Log, Config;

CONST
  mid = Cfg.midFiles;

  cntMax = Cfg.cnt;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. cntMax ];

  TFile = POINTER TO TFileDesc;
  TFileDesc = RECORD
    i: TIdx;          (* 0 <==> free *)
    w: BOOLEAN;
    r: UFTP.TResID;
    b: FIO.File;
    t: UFTP.TPosDesc;  (* tgt *)
  END;
  TFilesDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TFileDesc;
  END;

  TPthStr = ARRAY [ 1 .. 2 * SIZE( UFTP.TResID ) ] OF CHAR;

VAR
  fil: TFilesDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrTgt( VAR b: SO.TBuffer; t: UFTP.TPosDesc );
VAR
  c: CHAR;
  s: MD5.TDgstStr;
BEGIN
  c := b.c.p;
  b.c.p := '0';
  SO.WrLngHex( b, t.l, 8 );
  SO.WrChar( b, ' ' );
  MD5.MkStr( s, t.d );
  SO.WrStr( b, s );
  b.c.p := c;
END WrTgt;

PROCEDURE CalcMD5( f: FIO.File; VAR d: MD5.TDgstDesc ): LONGCARD;
CONST
  len = 4 * 1024;
VAR
  b: ARRAY [1..len] OF BYTE;
  l: CARDINAL;
  c: MD5.TCtxDesc;
  p: LONGCARD;
BEGIN
  FIO.Seek( f, 0 );
  p := 0;
  MD5.Init( c );
  l := FIO.RdBin( f, b, len );
  WHILE l > 0 DO
    INC( p, UFTP.TFileLen( l ) );
    MD5.Update( c, b, l );
    l := FIO.RdBin( f, b, len );
  END;
  MD5.Final( c, d );
  IF p # FIO.GetPos( f ) THEN (* happens after last successful read (l<len) *) END;
  RETURN p;
END CalcMD5;

PROCEDURE CalcMD5x( f: FIO.File; n: LONGCARD; VAR d: MD5.TDgstDesc );
CONST
  len = 4 * 1024;
VAR
  b: ARRAY [1..len] OF BYTE;
  l, m: CARDINAL;
  c: MD5.TCtxDesc;
BEGIN
  FIO.Seek( f, 0 );
  MD5.Init( c );
  WHILE n > 0 DO
    l := len;
    IF LONGCARD( l ) > n THEN l := CARDINAL( n ) END;
    m := FIO.RdBin( f, b, l );
    IF m # l THEN Fatal( 'CalcMD5x read error' ) END;
    MD5.Update( c, b, m );
    DEC( n, LONGCARD( m ) );
  END;
  MD5.Final( c, d );
END CalcMD5x;


PROCEDURE MkRdPth( VAR x: ARRAY OF CHAR; r-: UFTP.TResID );
BEGIN
  Str.Copy( x, r );
END MkRdPth;

PROCEDURE MkWrPth( VAR x: ARRAY OF CHAR; r-: UFTP.TResID );
BEGIN
  Str.Copy( x, (* r *) 'out.bin' );
END MkWrPth;

PROCEDURE Alloc( w: BOOLEAN; r-: UFTP.TResID; b: FIO.File; t-: UFTP.TPosDesc ): TFile;
VAR
  i: TVldIdx;
  f: TFile;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Alloc idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' wrt ' );
    SO.WrCard( b, ORD( w ), 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, r );
    SO.WrStr( b, "'" );
    SO.WrStr( b, ' tgt ' );
    WrTgt( b, t );
    Message( s );
  END Show;

BEGIN
  IF fil.n >= cntMax THEN RETURN NIL END;
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    f := ADR( fil.a[i] );
    IF f^.i = 0 THEN
      f^.i := i;
      f^.w := w;
      f^.r := r;
      f^.b := b;
      f^.t := t;
      INC( fil.n );
      Show();
      RETURN f;
    ELSIF f^.i # i THEN
      Fatal( 'Alloc idx inconsistency' );
    END;
  END;
  Fatal( 'Alloc cnt inconsistency' );
END Alloc;

PROCEDURE Free( VAR f: TFile );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Free idx ' );
    SO.WrCard( b, f^.i, 1 );
    SO.WrStr( b, ' wrt ' );
    SO.WrCard( b, ORD( f^.w ), 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, f^.r );
    SO.WrStr( b, "'" );
    SO.WrStr( b, ' tgt ' );
    WrTgt( b, f^.t );
    Message( s );
  END Show;

BEGIN
  IF ( f = NIL ) OR ( f^.i = 0 ) THEN Fatal( 'Free invalid file' ) END;
  Show();
  Lib2.Fill( f^, 0 );
  DEC( fil.n );
  f := NIL;
END Free;


PROCEDURE OpenRd( r-: UFTP.TResID ): TFile;
VAR
  p: TPthStr;
  b: FIO.File;
  t: UFTP.TPosDesc;
  f: TFile;
BEGIN
  IF fil.n >= cntMax THEN RETURN NIL END;
  MkRdPth( p, r );
  IF ~ FIO.Exists( p ) THEN RETURN NIL END;
  b := FIO.OpenRead( p );
  IF b = FIOStd.inv THEN RETURN NIL END;
  t.l := CalcMD5( b, t.d );
  f := Alloc( FALSE, r, b, t );
  RETURN f;
END OpenRd;

PROCEDURE Read( f: TFile; p: LONGCARD; VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
BEGIN
  IF f^.w THEN Fatal( 'Read mode' ) END;
  IF l > SIZE( x ) THEN Fatal( 'Read len' ) END;
  FIO.Seek( f^.b, p );
  l := FIO.RdBin( f^.b, x, l );
  RETURN l;
END Read;


PROCEDURE OpenWr( r-: UFTP.TResID ): TFile;
VAR
  p: TPthStr;
  b: FIO.File;
  t: UFTP.TPosDesc;
  f: TFile;
BEGIN
  IF fil.n >= cntMax THEN RETURN NIL END;
  MkWrPth( p, r );
  b := FIO.Create( p );
  IF b = FIOStd.inv THEN RETURN NIL END;
  Lib2.Fill( t, 0 );
  f := Alloc( TRUE, r, b, t );
  RETURN f;
END OpenWr;

PROCEDURE SetTgt( f: TFile; t-: UFTP.TPosDesc );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'SetTgt idx ' );
    SO.WrCard( b, f^.i, 1 );
    SO.WrStr( b, ' wrt ' );
    SO.WrCard( b, ORD( f^.w ), 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, f^.r );
    SO.WrStr( b, "'" );
    SO.WrStr( b, ' tgt ' );
    WrTgt( b, f^.t );
    Message( s );
  END Show;

BEGIN
  IF ~ f^.w THEN Fatal( 'SetTgt mode' ) END;
  f^.t := t;
  Show();
END SetTgt;

PROCEDURE Write( f: TFile; p: LONGCARD; x-: ARRAY OF BYTE; l: CARDINAL );
BEGIN
  IF ~ f^.w THEN Fatal( 'Write mode' ) END;
  IF l > SIZE( x ) THEN Fatal( 'Write len' ) END;
  FIO.Seek( f^.b, p );
  FIO.WrBin( f^.b, x, l );
END Write;

PROCEDURE Trunc( f: TFile; p: LONGCARD );
BEGIN
  IF ~ f^.w THEN Fatal( 'Trunc mode' ) END;
  FIO.Seek( f^.b, p );
  FIO.Truncate( f^.b );
END Trunc;


PROCEDURE GetRID( f: TFile; VAR x: UFTP.TResID );
BEGIN
  x := f^.r;
END GetRID;

PROCEDURE GetTgt( f: TFile; VAR x: UFTP.TPosDesc );
BEGIN
  x := f^.t;
END GetTgt;

PROCEDURE GetPos( f: TFile; p: LONGCARD; VAR x: UFTP.TPosDesc );
BEGIN
  IF f^.w THEN
    x.l := CalcMD5( f^.b, x.d );
  ELSE
    IF p >= f^.t.l THEN x := f^.t; RETURN END;
    x.l := p;
    CalcMD5x( f^.b, p, x.d );
  END;
END GetPos;


PROCEDURE Close( VAR f: TFile );
BEGIN
  FIO.Close( f^.b );
  Free( f );
END Close;


PROCEDURE CloseAll();
VAR
  i: TVldIdx;
  f: TFile;
BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    f := ADR( fil.a[i] );
    IF f^.i > 0 THEN Close( f ) END;
  END;
END CloseAll;

VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule' );
  Message( s );
  CloseAll();
  oldTerm();
END DoneModule;

BEGIN
  Log.Register( mid, 'Files' );
  Lib2.Fill( fil, 0 );

  Lib3.Terminate( DoneModule, oldTerm );
END Files.
