(* UFTP senders *)
(* solyga@gmx.de 2022-02-24 *)
IMPLEMENTATION MODULE Sndrs;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib2, Lib3;
IMPORT UFTP, MD5, Cfg, SO, Log, Config, Files;

CONST
  period = 1;  (* ms *)
  mid = Cfg.midSndrs;

  cntMax = Cfg.cnt;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. MAX( TIdx ) ];

CONST
  cntMaxReq = 64;

TYPE
  TReqsDesc = RECORD
    n,
    r,
    w: CARDINAL;
    a: ARRAY [ 0 .. cntMaxReq - 1 ] OF UFTP.TFgmIdx;
    m: UFTP.TFgmIdx;  (* fgm cnt *)
  END;

TYPE
  TSndr = POINTER TO TSndrDesc;
  TSndrDesc = RECORD
    i: TIdx;           (* 0 <==> free *)
    l: UFTP.TFgmLen;
    f: Files.TFile;
    r: TReqsDesc;
  END;

  TSndrsDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TSndrDesc;
  END;

VAR
  snd: TSndrsDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* don't remove this before having checked return code implications *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE Alloc( l: UFTP.TFgmLen; f: Files.TFile ): TSndr;
VAR
  i: TVldIdx;
  s: TSndr;
  t: UFTP.TPosDesc;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Alloc idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' fgl ' );
    SO.WrCard( b, l, 1 );
    Message( s );
  END Show;

BEGIN
  IF ( snd.n >= cntMax ) OR ( l = 0 ) OR ( l > UFTP.lenMaxFgm ) OR ( f = Files.fNull ) THEN RETURN NIL END;
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    s := ADR( snd.a[i] );
    IF s^.i = 0 THEN
      Lib2.Fill( s^, 0 );
      s^.i := i;
      s^.l := l;
      s^.f := f;
      Files.GetTgt( f, t );
      s^.r.m := t.l DIV l + ORD( t.l MOD l > 0 );
      INC( snd.n );
      Show();
      RETURN s;
    END;
  END;
  Fatal( 'Alloc cnt inconsistency' );
END Alloc;

PROCEDURE GetFgl( s: TSndr; VAR l: UFTP.TFgmLen );
BEGIN
  l := s^.l;
END GetFgl;

PROCEDURE Receive( s: TSndr; i: UFTP.TFgmIdx; x: UFTP.TFgmSet );

  PROCEDURE Add( VAR r: TReqsDesc; i: UFTP.TFgmIdx );
  BEGIN
    IF ( r.n >= cntMaxReq ) OR ( i >= r.m ) THEN RETURN END;
    r.a[ r.w ] := i;
    r.w := ( r.w + 1 ) MOD cntMaxReq;
    INC( r.n );
  END Add;

VAR
  j: UFTP.TRelFgmIdx;
BEGIN
  FOR j := MIN( UFTP.TRelFgmIdx ) TO MAX( UFTP.TRelFgmIdx ) DO
    IF j IN x THEN Add( s^.r, i + j ) END;
  END;
END Receive;

PROCEDURE Update( s: TSndr; n: CARDINAL; p: TSndProc; r: TRef );

  PROCEDURE Get( VAR r: TReqsDesc; VAR i: UFTP.TFgmIdx );
  BEGIN
    i := r.a[ r.r ];
    r.r := ( r.r + 1 ) MOD cntMaxReq;
    DEC( r.n );
  END Get;

VAR
  i: UFTP.TFgmIdx;
  l: CARDINAL;
  b: ARRAY [1..UFTP.lenMaxFgm] OF BYTE;
BEGIN
  WHILE ( n > 0 ) & ( s^.r.n > 0 ) DO
    Get( s^.r, i );
    l := s^.l;
    l := Files.Read( s^.f, i * l, b, l );
    p( r, i, b, l );
    DEC( n );
  END;
END Update;

PROCEDURE Free( VAR s: TSndr );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    t: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, t );
    SO.WrStr( b, 'Free idx ' );
    SO.WrCard( b, s^.i, 1 );
    SO.WrStr( b, ' fgl ' );
    SO.WrCard( b, s^.l, 1 );
    Message( t );
  END Show;

BEGIN
  Show();
  Lib2.Fill( s^, 0 );
  s := NIL;
END Free;


PROCEDURE Show();
CONST
  ena = TRUE;
VAR
  b: SO.TBuffer;
  s: ARRAY [1..128] OF CHAR;
BEGIN
  IF ~ ena THEN RETURN END;
  SO.Init( b, s );
  SO.WrStr( b, 'InitModule' );
  SO.WrStr( b, ' cntMax ' );
  SO.WrCard( b, cntMax, 1 );
  SO.WrStr( b, ', SIZE( TSndrDesc ) = ' );
  SO.WrCard( b, SIZE( TSndrDesc ), 1 );
  Message( s );
END Show;

BEGIN
  Log.Register( mid, 'Sndrs' );
  Lib2.Fill( snd, 0 );
  Show();
END Sndrs.
