(* callback timers *)
(* solyga@gmx.de, 2019-03-18 *)
IMPLEMENTATION MODULE Tmrs;

FROM SYSTEM IMPORT ADR;
IMPORT Tmr0, Lib2;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. cntMax ];

  TTmrDesc = RECORD
    i: TIdx;
    v: TTicks;
    n: TTicks;  (* down-counter, 0 = off *)
    m: TTicks;  (* reload, 0 = single shot *)
    h: THnd;
    r: TRef;
  END;
  TTimer = POINTER TO TTmrDesc;

  TTmrArr = ARRAY TVldIdx OF TTmrDesc;

VAR
  cnt: TCnt;
  arr: TTmrArr;


PROCEDURE New(): TTimer;
VAR
  t: TTimer;
BEGIN
  INC( cnt );
  t := ADR( arr[ cnt ] );
  Lib2.Fill( t^, 0 );
  t^.i := cnt;
  t^.h := NULLPROC;
  RETURN t;
END New;

PROCEDURE Set( t: TTimer; n, m: TTicks );
BEGIN
  t^.n := n;
  t^.m := m;
END Set;

PROCEDURE Val( t: TTimer ): TTicks;
BEGIN
  RETURN t^.v;
END Val;

PROCEDURE SetHnd( t: TTimer; h: THnd; r: TRef );
BEGIN
  t^.h := h;
  t^.r := r;
END SetHnd;


PROCEDURE Upd( VAR t: TTmrDesc; d: TTicks );
VAR
  n, k: TTicks;
BEGIN
  n := t.n;
  IF n = 0 THEN RETURN END;
  IF n > d THEN t.n := n - d; RETURN END;
  (* at least one hit *)
  k := 1;
  DEC( d, n );
  n := t.m;
  IF ( n > 0 ) & ( d > 0 ) THEN
    INC( k, d DIV n );
    DEC( n, d MOD n );
  END;
  t.n := n;
  Lib2.IncL( t.v, k );
  IF t.h # NULLPROC THEN t.h( k, t.r ) END;
END Upd;

PROCEDURE Trigger( d: TTicks );
VAR
  n: TCnt;
BEGIN
  IF d <= 0 THEN RETURN END;
  n := cnt;
  WHILE n > 0 DO Upd( arr[ n ], d ); DEC( n ); END;
END Trigger;


BEGIN
  cnt := 0;
END Tmrs.
